import asyncio
import functools

import hyundai_kia_connect_api as api

from .. import fhem, generic, utils


class kia_hyundai(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "update_interval": {
                "default": 30,
                "format": "int",
                "help": "Change interval in minutes, default is 30.",
            }
        }
        await self.set_attr_config(attr_config)

        set_config = {
            "lock": {},
            "unlock": {},
            "start_climate": {
                "args": ["set_temp", "duration", "defrost", "climate", "heating"],
                "help": "Params: Temperature Duration Defrost(on/off) Climate(on/off) Heating(on/off)",
                "params": {
                    "set_temp": {"default": 21, "format": "int"},
                    "duration": {"default": 5, "format": "int"},
                    "defrost": {"default": False, "format": "bool"},
                    "climate": {"default": True, "format": "bool"},
                    "heating": {"default": False, "format": "bool"},
                },
            },
            "stop_climate": {},
            "start_charge": {},
            "stop_charge": {},
            "charge_limits": {
                "args": ["ac_limit", "dc_limit"],
                "help": "Params AC Limit DC Limit",
                "params": {
                    "ac_limit": {"default": 90, "format": "int"},
                    "dc_limit": {"default": 90, "format": "int"},
                },
            },
            "update_data": {},
        }
        await self.set_set_config(set_config)

        if len(args) != 8:
            return (
                "Usage: define my_car fhempy kia_hyundai "
                + "USERNAME PASSWORD PIN CAR_BRAND REGION_CODE"
            )
        self.username = args[3]
        self.password = args[4]
        self.pin = args[5]
        self.car_brand = [
            k for k, v in api.const.BRANDS.items() if v.lower() == args[6].lower()
        ][0]
        if args[7].lower() == "eu":
            args[7] = "Europe"
        self.region_code = [
            k for k, v in api.const.REGIONS.items() if v.lower() == args[7].lower()
        ][0]
        self.create_async_task(self.connect_to_car())

    async def connect_to_car(self):
        self.vm = api.VehicleManager(
            self.region_code,
            self.car_brand,
            self.username,
            self.password,
            self.pin,
            geocode_api_enable=True,
            geocode_api_use_email=True,
        )
        while True:
            try:
                await utils.run_blocking(
                    functools.partial(self.vm.check_and_refresh_token)
                )
                self.create_async_task(self.update_loop())
                break
            except Exception:
                self.logger.exception("Failed to login to server, retry in 120s")
                await asyncio.sleep(120)

    async def update_loop(self):
        while True:
            await self.update_once()
            await asyncio.sleep(self._attr_update_interval * 60)

    async def update_once(self):
        try:
            await utils.run_blocking(functools.partial(self.vm.check_and_refresh_token))
            await utils.run_blocking(
                functools.partial(self.vm.update_all_vehicles_with_cached_state)
            )
            await self.update_readings()
        except Exception:
            self.logger.exception("Failed to update car data")
            await fhem.readingsSingleUpdate(self.hash, "state", "error", 1)

    async def update_readings(self):
        # use only first vehicle
        self.vehicle = self.vm.vehicles[list(self.vm.vehicles)[0]]
        flat_json = utils.flatten_json(self.vehicle.data)
        await fhem.readingsBeginUpdate(self.hash)
        try:
            for name in flat_json:
                await fhem.readingsBulkUpdate(self.hash, name, flat_json[name])
            await fhem.readingsBulkUpdate(
                self.hash, "geolocation", self.vehicle.geocode[0]
            )
            await fhem.readingsBulkUpdate(self.hash, "state", "online")
        except Exception:
            self.logger.exception("Failed to update readings")
        await fhem.readingsEndUpdate(self.hash, 1)

    # Set functions in format: set_NAMEOFSETFUNCTION(self, hash, params)
    async def set_update_data(self, hash, params):
        self.create_async_task(self.update_once())

    async def set_lock(self, hash, params):
        self.create_async_task(self.execute_command(self.vm.lock, self.vehicle.id))

    async def set_unlock(self, hash, params):
        self.create_async_task(self.execute_command(self.vm.unlock, self.vehicle.id))

    async def set_start_climate(self, hash, params):
        self.create_async_task(
            self.execute_command(
                self.vm.start_climate,
                self.vehicle.id,
                params["set_temp"],
                params["duration"],
                params["defrost"],
                params["climate"],
                params["heating"],
            )
        )

    async def set_stop_climate(self, hash, params):
        self.create_async_task(
            self.execute_command(self.vm.stop_climate, self.vehicle.id)
        )

    async def set_start_charge(self, hash, params):
        self.create_async_task(
            self.execute_command(self.vm.start_charge, self.vehicle.id)
        )

    async def set_stop_charge(self, hash, params):
        self.create_async_task(
            self.execute_command(self.vm.stop_charge, self.vehicle.id)
        )

    async def set_charge_limits(self, hash, params):
        utils.run_blocking_task(
            functools.partial(
                self.vm.set_charge_limits,
                self.vehicle.id,
                params["ac_limit"],
                params["dc_limit"],
            )
        )

    async def execute_command(self, command, *kwargs):
        # check token
        await utils.run_blocking(functools.partial(self.vm.check_and_refresh_token))

        # execute function
        await utils.run_blocking(functools.partial(command, *kwargs))



import asyncio
import functools
import json
import os
import shutil
import signal
import subprocess

import netifaces
from fhempy.lib.generic import FhemModule
from git import Repo

from .. import fhem, utils


class zigbee2mqtt(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.proc = None
        self.check_process_task = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        self._set_list = {"start": {}, "stop": {}, "restart": {}, "update": {}}
        await self.set_set_config(self._set_list)
        attr_conf = {"disable": {"options": "0,1", "default": "0", "format": "int"}}
        await self.set_attr_config(attr_conf)

        if self._attr_disable == 1:
            return

        self.create_async_task(self.run_z2m())

    async def run_z2m(self):
        res = await self.check_node_installation()
        if res is False:
            return

        res = await self.check_npm_installation()
        if res is False:
            return

        res = await self.install_z2m()
        if res is False:
            return

        await self.start_process()

    async def update_z2m(self):
        try:
            await fhem.readingsSingleUpdate(
                self.hash, "update", "started, may take a few minutes", 1
            )
            current_directory = os.getcwd()
            z2m_directory = os.path.join(current_directory, r".fhempy/zigbee2mqtt")

            await self.stop_process()
            try:
                await utils.run_blocking(
                    functools.partial(
                        shutil.copytree,
                        z2m_directory + "/data",
                        z2m_directory + "/data-backup",
                    )
                )
            except FileExistsError:
                shutil.rmtree(z2m_directory + "/data-backup")
                await utils.run_blocking(
                    functools.partial(
                        shutil.copytree,
                        z2m_directory + "/data",
                        z2m_directory + "/data-backup",
                    )
                )

            rep = Repo(z2m_directory)
            o = rep.remotes.origin
            await utils.run_blocking(functools.partial(o.pull))

            await utils.run_blocking(
                functools.partial(subprocess.call, ["npm", "ci"], cwd=z2m_directory)
            )

            source_folder = z2m_directory + "/data-backup"
            destination_folder = z2m_directory + "/data"
            for file in os.listdir(source_folder):
                file_path = os.path.join(source_folder, file)
                if os.path.isfile(file_path):
                    await utils.run_blocking(
                        functools.partial(shutil.copy, file_path, destination_folder)
                    )

            await fhem.readingsSingleUpdate(self.hash, "update", "successful", 1)
            await self.start_process()
        except Exception:
            self.logger.exception("Failed to update")
            await fhem.readingsSingleUpdate(
                self.hash, "update", "failed to update, check log", 1
            )

    async def install_z2m(self):
        GIT_URL = "https://github.com/Koenkk/zigbee2mqtt.git"

        current_directory = os.getcwd()
        install_directory = os.path.join(current_directory, r".fhempy")
        if not os.path.exists(install_directory):
            os.makedirs(install_directory)
        z2m_directory = os.path.join(current_directory, r".fhempy/zigbee2mqtt")
        clone_needed = True
        if not os.path.exists(z2m_directory):
            os.makedirs(z2m_directory)
        else:
            try:
                Repo(z2m_directory)
                clone_needed = False
            except Exception:
                pass

        try:
            if clone_needed:
                await fhem.readingsSingleUpdate(
                    self.hash,
                    "installation",
                    "downloading zigbee2mqtt",
                    1,
                )
                await utils.run_blocking(
                    functools.partial(Repo.clone_from, GIT_URL, z2m_directory)
                )

            if (
                await fhem.ReadingsVal(self.hash["NAME"], "installation", "nok")
                != "successful"
            ):
                await fhem.readingsSingleUpdate(
                    self.hash,
                    "installation",
                    "installing dependencies, might take some minutes",
                    1,
                )
                await utils.run_blocking(
                    functools.partial(subprocess.call, ["npm", "ci"], cwd=z2m_directory)
                )

                z2m_conf = (
                    "homeassistant: false\n"
                    "permit_join: false\n"
                    "mqtt:\n"
                    "  base_topic: zigbee2mqtt\n"
                    "  server: 'mqtt://localhost'\n"
                    "  client_id: 'zigbee_pi'\n"
                    "serial:\n"
                    "  port: /dev/ttyUSB0\n"
                    "frontend:\n"
                    "  port: 8080\n"
                )
                with open(z2m_directory + "/data/configuration.yaml", "w") as file:
                    file.writelines(z2m_conf)

                await self.create_weblink()
                await fhem.readingsSingleUpdate(
                    self.hash, "installation", "successful", 1
                )
        except Exception:
            self.logger.exception("Failed to clone repo")
            await fhem.readingsSingleUpdate(
                self.hash,
                "state",
                "failed to install, check fhempy log",
                1,
            )
            return False

        await fhem.readingsSingleUpdate(self.hash, "state", "ready", 1)

        return True

    async def check_node_installation(self):
        try:
            ver = await utils.run_blocking(
                functools.partial(subprocess.check_output, ["node", "--version"])
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "node", ver.decode("ascii").rstrip(), 1
            )
            return True
        except Exception:
            await fhem.readingsSingleUpdate(
                self.hash, "node", "nodejs installation missing", 1
            )
            return False

    async def check_npm_installation(self):
        try:
            ver = await utils.run_blocking(
                functools.partial(subprocess.check_output, ["npm", "--version"])
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "npm", ver.decode("ascii").rstrip(), 1
            )
            return True
        except Exception:
            await fhem.readingsSingleUpdate(
                self.hash, "npm", "npm installation missing", 1
            )
            return False

    async def start_process(self):
        current_directory = os.getcwd()
        z2m_directory = os.path.join(current_directory, r".fhempy/zigbee2mqtt")

        # Load the package.json file
        with open(z2m_directory + "/package.json", "r") as f:
            package_data = json.load(f)

        # Extract the version number
        version = package_data["version"]
        await fhem.readingsSingleUpdate(self.hash, "z2m_version", version, 1)

        try:
            self.proc = subprocess.Popen(["node", "./index.js"], cwd=z2m_directory)
            await fhem.readingsSingleUpdate(self.hash, "state", "running", 1)
            if self.check_process_task is None:
                self.check_process_task = self.create_async_task(self.check_process())
        except Exception:
            self.logger.exception("Failed to start zigbee2mqtt with npm start")

    async def check_process(self):
        while True:
            try:
                if self.proc is not None:
                    poll = self.proc.poll()
                    if poll is None:
                        await fhem.readingsSingleUpdateIfChanged(
                            self.hash, "state", "running", 1
                        )
                    elif poll != 0:
                        await fhem.readingsSingleUpdate(
                            self.hash, "state", "error, check log file", 1
                        )
                        await asyncio.sleep(10)
                        await self.start_process()
                    elif poll == 0:
                        await fhem.readingsSingleUpdate(
                            self.hash, "state", "stopped", 1
                        )
                        await asyncio.sleep(10)
                        await self.start_process()
                await asyncio.sleep(10)
            except asyncio.CancelledError:
                return

    async def stop_process(self):
        if self.check_process_task:
            self.cancel_async_task(self.check_process_task)
            self.check_process_task = None
        if self.proc:
            await fhem.readingsSingleUpdate(self.hash, "state", "stopping", 1)
            self.proc.send_signal(signal.SIGINT)

            stop_tries = 0
            # give zigbee2mqtt some time to stop
            await asyncio.sleep(15)
            while self.proc.poll is None and stop_tries < 5:
                await asyncio.sleep(5)
                self.proc.kill()
                stop_tries += 1

            if self.proc.poll is None:
                self.logger.error("Failed to stop zigbee2mqtt process")
                await fhem.readingsSingleUpdate(self.hash, "state", "failed to stop", 1)
            else:
                # this should never block, as poll says process finished already
                # this should prevent zombie processes
                self.proc.wait(0.1)
                self.proc = None
                await fhem.readingsSingleUpdate(self.hash, "state", "stopped", 1)

    async def create_weblink(self):
        ip_list = [
            netifaces.ifaddresses(iface)[netifaces.AF_INET][0]["addr"]
            for iface in netifaces.interfaces()
            if netifaces.AF_INET in netifaces.ifaddresses(iface) and iface != "lo"
        ]
        await fhem.CommandDefine(
            self.hash, "z2m_frontend weblink iframe http://" + ip_list[0] + ":8080/"
        )
        await fhem.CommandAttr(
            self.hash,
            (
                "z2m_frontend htmlattr width='100%' height='90%' "
                "frameborder='0' marginheight='0' marginwidth='0'"
            ),
        )
        await fhem.CommandAttr(self.hash, "z2m_frontend room Zigbee2MQTT")

    # FHEM FUNCTION
    async def Undefine(self, hash):
        await self.stop_process()
        return await super().Undefine(hash)

    async def set_attr_disable(self, hash):
        if self._attr_disable == 0:
            self.create_async_task(self.start_process())
        else:
            self.create_async_task(self.stop_process())

    async def set_start(self, hash, params):
        self.create_async_task(self._restart())
        return ""

    async def set_stop(self, hash, params):
        self.create_async_task(self.stop_process())
        return ""

    async def set_restart(self, hash, params):
        self.create_async_task(self._restart())

    async def _restart(self):
        await self.stop_process()
        # wait for z2m to release hardware connection
        await asyncio.sleep(10)
        await self.start_process()

    async def set_update(self, hash, params):
        self.create_async_task(self.update_z2m())



"""Helper for ble_monitor."""
import logging
import re
import voluptuous as vol
from uuid import UUID
from typing import Optional, Any
from .const import MAC_REGEX, AES128KEY24_REGEX, AES128KEY32_REGEX, CONF_UUID, CONF_MAC

_LOGGER = logging.getLogger(__name__)


def identifier_normalize(value: str) -> str:
    if validate_uuid(value):
        return str(UUID(value))

    if ":" in value:
        return value

    return ":".join(value[i : i + 2] for i in range(0, len(value), 2))


def detect_conf_type(value: str) -> str:
    return CONF_UUID if validate_uuid(value) else CONF_MAC


def dict_get_or(
    data: dict, first: str = CONF_UUID, second: str = CONF_MAC
) -> Optional[str]:
    key = dict_get_key_or(data, first, second)

    return data[key] if key in data else None


def dict_get_or_normalize(
    data: dict, first: str = CONF_UUID, second: str = CONF_MAC
) -> Optional[str]:
    key = dict_get_key_or(data, first, second)

    return identifier_normalize(data[key]) if key in data else None


def dict_get_or_clean(
    data: dict, first: str = CONF_UUID, second: str = CONF_MAC
) -> str:
    return identifier_clean(dict_get_or(data, first, second))


def dict_get_key_or(data: dict, first: str = CONF_UUID, second: str = CONF_MAC) -> str:
    return first if first in data and data[first] else second


def identifier_clean(value: str) -> str:
    return value.replace("-", "").replace(":", "")


def validate_mac(value: str) -> bool:
    """Mac validation."""
    return _validate_regex(value, MAC_REGEX)


def validate_uuid(value: str) -> bool:
    """UUID validation."""
    try:
        config_validation_uuid(value)

        return True
    except vol.Invalid:
        return False


def validate_key(value: str) -> bool:
    """Key validation."""
    if not value or value == "-":
        return True

    if not _validate_regex(value, AES128KEY24_REGEX):
        if not _validate_regex(value, AES128KEY32_REGEX):
            return False

    return True


def _validate_regex(value: str, regex: str) -> bool:
    """Validate that the value is a string that matches a regex."""
    compiled = re.compile(regex)
    if not compiled.match(value):
        return False
    return True


def config_validation_uuid(value: Any) -> str:
    try:
        result = str(UUID(value))
    except (ValueError, AttributeError, TypeError) as error:
        raise vol.Invalid("Invalid UUID", error_message=str(error))

    if identifier_clean(result).lower() != identifier_clean(value).lower():
        raise vol.Invalid("Invalid UUID")

    return result



import functools

from fhempy.lib.generic import FhemModule
from micloud import MiCloud

from .. import fhem, utils


class xiaomi_tokens(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._username = None
        self._password = None
        self._country = ["de", "cn", "sg"]
        self._all_devices = {}

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        self._set_list_conf = {
            "username": {"args": ["username"]},
            "password": {"args": ["password"]},
            "get_tokens": {},
        }
        await self.set_set_config(self._set_list_conf)
        
        await fhem.readingsSingleUpdateIfChanged(hash, "state", "active", 1)

        self._uniqueid = await fhem.getUniqueId(self.hash)
        self._enc_username = await fhem.ReadingsVal(
            self.hash["NAME"], "xiaomi_username", ""
        )
        self._enc_username = self._enc_username.replace("\\n", "")
        if self._enc_username != "":
            self._username = utils.decrypt_string(self._enc_username, self._uniqueid)

        self._enc_password = await fhem.ReadingsVal(
            self.hash["NAME"], "xiaomi_password", ""
        )
        self._enc_password = self._enc_password.replace("\\n", "")
        if self._enc_password != "":
            self._password = utils.decrypt_string(self._enc_password, self._uniqueid)

        # retrieve tokens
        if self._username and self._password:
            self.create_async_task(self.obtain_tokens())
        return

    async def set_username(self, hash, params):
        self._username = params["username"]
        self._enc_username = utils.encrypt_string(self._username, self._uniqueid)
        await fhem.readingsSingleUpdateIfChanged(
            hash, "xiaomi_username", self._enc_username, 1
        )
        return ""

    async def set_password(self, hash, params):
        self._password = params["password"]
        self._enc_password = utils.encrypt_string(self._password, self._uniqueid)
        await fhem.readingsSingleUpdateIfChanged(
            hash, "xiaomi_password", self._enc_password, 1
        )
        return ""

    async def set_get_tokens(self, hash, params):
        if self._username and self._password:
            self.create_async_task(self.obtain_tokens())
        else:
            return "Please set username & password first!"

    async def set_create_miio_device(self, hash, params):
        arr = params["dev"].split("_")
        did = arr[0]
        country = arr[1]
        model = self._all_devices[did + country]["model"]
        ip = self._all_devices[did + country]["localip"]
        token = self._all_devices[did + country]["token"]
        if "vacuum" in model:
            self.create_async_task(
                fhem.CommandDefine(
                    self.hash,
                    f"miio_vacuum_{did} fhempy miio vacuum {ip} {token}",
                )
            )
        elif "viomi" in model:
            self.create_async_task(
                fhem.CommandDefine(
                    self.hash,
                    f"miio_vacuum_{did} fhempy miio viomivacuum {ip} {token}",
                )
            )
        elif "chuangmi" in model or "camera" in model:
            self.create_async_task(
                fhem.CommandDefine(
                    self.hash,
                    f"miio_camera_{did} fhempy miio chuangmicamera {ip} {token}",
                )
            )
        else:
            self.create_async_task(
                fhem.CommandDefine(
                    self.hash,
                    f"miio_device_{did} fhempy miio device {ip} {token}",
                )
            )

    async def set_create_gateway3_device(self, hash, params):
        arr = params["dev"].split("_")
        did = arr[0]
        country = arr[1]
        ip = self._all_devices[did + country]["localip"]
        token = self._all_devices[did + country]["token"]
        if ip != "" and token != "":
            self.create_async_task(
                fhem.CommandDefine(
                    self.hash,
                    f"xiaomigw3_{did} fhempy xiaomi_gateway3 {ip} {token}",
                )
            )

    async def obtain_tokens(self):
        try:
            await utils.run_blocking(functools.partial(self.thread_get_tokens))
        except Exception as ex:
            await self.logger.error(f"Failed to get tokens: {ex}")
            await fhem.readingsSingleUpdate(self.hash, "state", f"{ex}", 1)
            return

        self._miio_devices = []
        self._xiaomigw3_devices = []
        await fhem.readingsBeginUpdate(self.hash)
        for dev in self._device_list:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, dev["did"] + "_" + dev["country"] + "_name", dev["name"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, dev["did"] + "_" + dev["country"] + "_token", dev["token"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, dev["did"] + "_" + dev["country"] + "_model", dev["model"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, dev["did"] + "_" + dev["country"] + "_ip", dev["localip"]
            )
            self._all_devices[dev["did"] + dev["country"]] = dev
            if dev["did"][0:3] != "blt" and dev["localip"] != "":
                self._miio_devices.append(
                    dev["did"]
                    + "_"
                    + dev["country"]
                    + "_("
                    + dev["name"].replace(" ", "_")
                    + ")"
                )
                if dev["model"] == "lumi.gateway.mgl03":
                    self._xiaomigw3_devices.append(
                        dev["did"]
                        + "_"
                        + dev["country"]
                        + "_("
                        + dev["name"].replace(" ", "_")
                        + ")"
                    )
        await fhem.readingsEndUpdate(self.hash, 1)

        self._set_list_conf["create_miio_device"] = {
            "args": ["dev"],
            "help": "Creates a fhempy miio device to control your Xiaomi device.",
            "options": ",".join(self._miio_devices),
        }
        self._set_list_conf["create_gateway3_device"] = {
            "args": ["dev"],
            "help": "Creates a fhempy xiaomi_gateway3 device and devices for all connected devices.",
            "options": ",".join(self._xiaomigw3_devices),
        }
        await self.set_set_config(self._set_list_conf)

    def thread_get_tokens(self):
        self._device_list = []
        mc = MiCloud(self._username, self._password)
        if mc.login():
            for country in self._country:
                dev_list = mc.get_devices(country=country)
                for dev in dev_list:
                    dev["country"] = country
                    self._device_list.append(dev)
        else:
            raise Exception("Login failed, please check username/password!")



import asyncio
import functools
import json

import fhempy.lib.fhem as fhem
import fhempy.lib.utils as utils
import tuya_iot
from tuya_connector import TuyaCloudPulsarTopic, TuyaOpenPulsar

from .const import TUYA_ENDPOINT


class tuya_cloud_setup:
    def __init__(self, logger, fhemdevice):
        self.logger = logger
        self.fhemdev = fhemdevice
        self.hash = fhemdevice.hash
        self._t_devicelist = []
        self._ready = False

    async def Define(self, hash, args, argsh):
        self._t_apikey = args[4]
        self._t_apisecret = args[5]
        self._t_username = args[6]
        self._t_password = args[7]
        if len(args) > 8:
            self._t_apptype = args[8]
        else:
            self._t_apptype = "smartlife"

        if len(args) > 9:
            self._t_region = args[9]
        else:
            self._t_region = "Europe"
        self.hash["DEVICEID"] = 0

        self.fhemdev.create_async_task(self.run_setup())

    async def Undefine(self, hash):
        return

    def _get_region_url(self, region) -> str:
        for url in TUYA_ENDPOINT:
            if TUYA_ENDPOINT[url] == region:
                return url

    def _get_pulsar_endpoint(self, region):
        return "wss://mqe.tuya" + self._get_countrycode(self._t_region) + ".com:8285/"

    def _get_countrycode(self, region) -> str:
        if region == "Europe":
            return "eu"
        else:
            return "cn"

    async def run_setup(self):
        await fhem.readingsSingleUpdate(self.hash, "state", "connecting", 1)
        if await self._init_tuya_sdk() is True:
            self._ready = True
            await fhem.readingsSingleUpdate(self.hash, "state", "connected", 1)
            await self._init_devices()

    async def _init_tuya_openpulsar(self):
        self.tuya_pulsar = TuyaOpenPulsar(
            self._t_apikey,
            self._t_apisecret,
            self._get_pulsar_endpoint(self._t_region),
            TuyaCloudPulsarTopic.PROD,
        )
        self.tuya_pulsar.start()

        # check if pulsar is working
        await asyncio.sleep(2)
        pulsar_connected = False
        try:
            pulsar_connected = self.tuya_pulsar.ws_app.sock.status == 101
        except Exception:
            pass

        if pulsar_connected is False:
            self.logger.error(
                "Please activate OpenPulsar: "
                + "https://developer.tuya.com/en/docs/iot/subscribe-mq?id=Kavqcrvckbh9h"
            )
            # pulsar not working
            self.tuya_pulsar.stop()
        else:
            self.logger.info("Tuya Open Pulsar connected")

    async def _init_tuya_sdk(self) -> bool:
        auth_type = tuya_iot.AuthType(0)
        api = tuya_iot.TuyaOpenAPI(
            self._get_region_url(self._t_region),
            self._t_apikey,
            self._t_apisecret,
            auth_type,
        )

        api.set_dev_channel("fhempy")

        response = (
            await utils.run_blocking(
                functools.partial(api.connect, self._t_username, self._t_password)
            )
            if auth_type == tuya_iot.AuthType.CUSTOM
            else await utils.run_blocking(
                functools.partial(
                    api.connect,
                    self._t_username,
                    self._t_password,
                    self._get_countrycode(self._t_region),
                    self._t_apptype,
                )
            )
        )
        if response.get("success", False) is False:
            if response.get("code", 0) == 2406:
                await fhem.readingsSingleUpdate(
                    self.hash, "state", "Tuya project too old, create new one", 1
                )
            else:
                await fhem.readingsSingleUpdate(
                    self.hash, "state", f"failed to login: {response}", 1
                )
                self.logger.error(f"Tuya login error response: {response}")
            return False

        await self._init_tuya_openpulsar()

        self.tuya_mq = tuya_iot.TuyaOpenMQ(api)
        self.tuya_mq.start()

        self.device_manager = tuya_iot.TuyaDeviceManager(api, self.tuya_mq)

        # Get device list
        self.home_manager = tuya_iot.TuyaHomeManager(
            api, self.tuya_mq, self.device_manager
        )
        await utils.run_blocking(
            functools.partial(self.home_manager.update_device_cache)
        )
        t_cloud_setup = self

        class DeviceListener(tuya_iot.TuyaDeviceListener):
            """Device Update Listener."""

            def __init__(self, logger) -> None:
                super().__init__()
                self.logger = logger

            def update_device(self, device: tuya_iot.TuyaDevice):
                self.logger.debug(f"update_device received for {device.id}")
                for dev in t_cloud_setup.tuya_devices:
                    if dev.id == device.id:
                        try:
                            asyncio.run_coroutine_threadsafe(
                                dev.update(device), t_cloud_setup.fhemdev.loop
                            )
                        except Exception:
                            self.logger.exception("Failed to update device")

            def add_device(self, device: tuya_iot.TuyaDevice):
                self.logger.info(f"add_device received for {device.id}")
                try:
                    asyncio.run_coroutine_threadsafe(
                        self.add_fhem_device(device), t_cloud_setup.fhemdev.loop
                    )
                except Exception:
                    self.logger.exception("Failed to add device")

            async def add_fhem_device(self, device: tuya_iot.TuyaDevice):
                await t_cloud_setup._create_fhem_device(device.name, device.id)
                try:
                    self.tuya_mq.stop()
                except Exception:
                    pass
                self.tuya_mq = tuya_iot.TuyaOpenMQ(
                    t_cloud_setup.device_manager.device_manager.api
                )
                self.tuya_mq.start()

                t_cloud_setup.device_manager.mq = self.tuya_mq
                self.tuya_mq.add_message_listener(
                    t_cloud_setup.device_manager.on_message
                )

            def remove_device(self, device_id: str):
                self.logger.info(f"remove_device received for {device_id}")

        __listener = DeviceListener(self.logger)
        self.device_manager.add_device_listener(__listener)

        def on_pulsar_message(msg):
            msg = json.loads(msg)
            status = msg.get("status", [])
            device = {"id": msg["devId"], "status": status}
            self.logger.debug(f"update_device received: {msg}")
            for dev in t_cloud_setup.tuya_devices:
                if dev.id == device["id"]:
                    try:
                        asyncio.run_coroutine_threadsafe(
                            dev.update_readings_arr(device["status"]),
                            t_cloud_setup.fhemdev.loop,
                        )
                    except Exception:
                        self.logger.exception("Failed to update device")

        self.tuya_pulsar.add_message_listener(on_pulsar_message)

        return True

    @property
    def ready(self):
        return self._ready

    def register_tuya_device(self, device):
        self._t_devicelist.append(device)

    def unregister_tuya_device(self, device):
        self._t_devicelist.remove(device)

    @property
    def tuya_devices(self):
        return self._t_devicelist

    async def _create_fhem_device(self, name, device_id):
        devalias = name
        devname = "tuya_cloud_" + device_id
        devname = utils.gen_fhemdev_name(devname)
        device_exists = await fhem.checkIfDeviceExists(
            self.hash, "FHEMPYTYPE", "tuya_cloud", "DEVICEID", device_id
        )

        if not device_exists:
            device_exists = await fhem.checkIfDeviceExists(
                self.hash, "FHEMPYTYPE", "tuya", "DEVICEID", device_id
            )

        if not device_exists:
            self.logger.info(
                (
                    f"create: {devname} fhempy tuya_cloud "
                    f"{self.hash['NAME']} {device_id}"
                )
            )
            # define each device (CommandDefine ... tuya_cloud_setup_dev deviceid
            await fhem.CommandDefine(
                self.hash,
                (f"{devname} fhempy tuya_cloud " f"{self.hash['NAME']} {device_id}"),
            )
            await fhem.CommandAttr(self.hash, f"{devname} alias {devalias}")
            # wait for FHEM to handle CommandDefine
            await asyncio.sleep(1)

    async def _init_devices(self):
        # wait for init to complete, otherwise devices might not be available yet
        while await fhem.init_done(self.hash) == 0:
            await asyncio.sleep(3)

        # retrieve devices from device_manager and create them
        for device_id in self.device_manager.device_map:
            await self._create_fhem_device(
                self.device_manager.device_map[device_id].name, device_id
            )
            await asyncio.sleep(1)

    async def send_commands(self, deviceid, commands):
        await utils.run_blocking(
            functools.partial(self.device_manager.send_commands, deviceid, commands)
        )



import re

MAIN_SCRIPT_REGEX = re.compile("^main\S+\.js$")
API_GATEWAY_TOKEN_REGEX = re.compile(
    r"b2cApiKey\:\s*\"([A-Za-z0-9\-_]+)\"", re.IGNORECASE
)

PAGE_URL = "https://smartmeter-web.wienernetze.at/"
API_URL_ALT = "https://service.wienernetze.at/sm/api/"
API_URL = "https://api.wstw.at/gateway/WN_SMART_METER_PORTAL_API_B2C/1.0/"
REDIRECT_URI = "https://smartmeter-web.wienernetze.at/"
API_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
AUTH_URL = "https://log.wien/auth/realms/logwien/protocol/openid-connect/"  # noqa


LOGIN_ARGS = {
    "client_id": "wn-smartmeter",
    "redirect_uri": REDIRECT_URI,
    "response_mode": "fragment",
    "response_type": "code",
    "scope": "openid",
    "nonce": "",
}


def build_access_token_args(**kwargs):
    args = {
        "grant_type": "authorization_code",
        "client_id": "wn-smartmeter",
        "redirect_uri": REDIRECT_URI,
    }
    args.update(**kwargs)
    return args


def build_verbrauchs_args(**kwargs):
    args = {
        "period": "DAY",
        "accumulate": False,
        "offset": 0,
        "dayViewResolution": "QUARTER-HOUR",
    }
    args.update(**kwargs)
    return args
